# Copyright (c) 2016-2022 Association of Universities for Research in Astronomy, Inc. (AURA)
# For license information see LICENSE or https://opensource.org/licenses/BSD-3-Clause

"""A small version of GPP data model"
"""
from .atom import *
from .constraints import *
from .group import *
from .ids import *
from .magnitude import *
from .observation import *
from .observationmode import *
from .program import *
from .qastate import *
from .resource import *
from .semester import *
from .site import *
from .target import *
from .timeallocation import *
from .timingwindow import *
from .too import *

# Type alias for a night index.
NightIndex = int

# Type alias for multiple night indices.
NightIndices = npt.NDArray[NightIndex]



# Copyright (c) 2016-2022 Association of Universities for Research in Astronomy, Inc. (AURA)
# For license information see LICENSE or https://opensource.org/licenses/BSD-3-Clause

# This has to be outside of Observation to avoid circular imports.
from enum import Enum


class ObservationMode(str, Enum):
    """Observation Mode

    Members:
        - UNKNOWN = 'unknown'
        - IMAGING = 'imaging'
        - LONGSLIT = 'longslit'
        - IFU = 'ifu'
        - MOS = 'mos'
        - XD = 'xd'
        - CORON = 'coron'
        - NRM = 'nrm'
    """
    # TODO: This is not stored anywhere and is only used temporarily in the atom code in the
    # TODO: OcsProgramExtractor. Should it be stored anywhere or is it only used in intermediate
    # TODO: calculations? It seems to depend on the instrument and FPU.

    UNKNOWN = 'unknown'
    IMAGING = 'imaging'
    LONGSLIT = 'longslit'
    IFU = 'ifu'
    MOS = 'mos'
    XD = 'xd'
    CORON = 'coron'
    NRM = 'nrm'



# Copyright (c) 2016-2022 Association of Universities for Research in Astronomy, Inc. (AURA)
# For license information see LICENSE or https://opensource.org/licenses/BSD-3-Clause

from abc import ABC
from datetime import timedelta
from typing import FrozenSet, Optional, Union

from astropy.time import Time  # type: ignore


class ObservatoryProperties(ABC):
    """Observatory-specific methods.

       These are not tied to other components or
       structures, and allow computations to be implemented in one place.

    """
    _properties: Optional[Union[None, 'ObservatoryProperties']] = None

    @staticmethod
    def set_properties(cls) -> None:
        """Set properties for an specific Observatory

        Raises:
            ValueError: Illegal properties value.

        """
        if not issubclass(cls, ObservatoryProperties):
            raise ValueError('Illegal properties value.')
        ObservatoryProperties._properties = cls()

    @staticmethod
    def _check_properties() -> None:
        """ Check if any properties are set

        Raises:
            ValueError: Properties have not been set.

        """
        if ObservatoryProperties._properties is None:
            raise ValueError('Properties have not been set.')

    @staticmethod
    def determine_standard_time(resources: FrozenSet,
                                wavelengths: FrozenSet[float],
                                modes: FrozenSet,
                                cal_length: int) -> Time:
        """Determine standard time for a specific Observatory

        Args:
            resources (FrozenSet): Set of Resources(instruments, mask, etc).
            wavelengths (FrozenSet[float]): An array of Wavelengths to be observed.
            modes (FrozenSet): The different modes of observation.
            cal_length (int): The length (in seconds) of a calibration.

        Returns:
            Time: Value(s) of standard time
        """
        if ObservatoryProperties._properties is not None:
            ObservatoryProperties._check_properties()
            return ObservatoryProperties._properties.determine_standard_time(
                resources,
                wavelengths,
                modes,
                cal_length
            )
        else:
            raise ValueError('Properties have not been set.')

    @staticmethod
    def is_instrument(resource) -> bool:  # type: ignore
        """Determine if the given resource is an instrument or not.

        Args:
            resource (Resource): An instrument.

        Returns:
            bool: True is the resource is an instrument of the Observatory, otherwise False.
        """
        if ObservatoryProperties._properties is not None:
            return ObservatoryProperties._properties.is_instrument(resource)
        else:
            raise ValueError('Properties have not been set.')

    @staticmethod
    def acquisition_time(resource, observation_mode) -> Optional[timedelta]:  # type: ignore
        """Given a resource, check if it is an instrument, and if so, lookup the
           acquisition time for the specified mode.

        Args:
            resource (Resource): A resource that should be an Instrument.
            observation_mode (ObservationMode): The observation mode to be used.

        Returns:
            Optional[timedelta]: The acquisition time for the instrument in that specific mode.
        """
        if ObservatoryProperties._properties is not None:
            return ObservatoryProperties._properties.acquisition_time(resource, observation_mode)
        else:
            raise ValueError('Properties have not been set.')



# Copyright (c) 2016-2022 Association of Universities for Research in Astronomy, Inc. (AURA)
# For license information see LICENSE or https://opensource.org/licenses/BSD-3-Clause

# Basic type aliases for usefulness.
from datetime import timedelta
from typing import Final, List, TypeVar, Union

import numpy.typing as npt
from astropy.time import Time  # type: ignore

T = TypeVar('T')

ScalarOrNDArray = Union[T, npt.NDArray[T]]
TimeScalarOrNDArray = Union[Time, npt.NDArray[float]]
ListOrNDArray = Union[List[T], npt.NDArray[T]]

ZeroTime: Final[timedelta] = timedelta()



# Copyright (c) 2016-2022 Association of Universities for Research in Astronomy, Inc. (AURA)
# For license information see LICENSE or https://opensource.org/licenses/BSD-3-Clause

from dataclasses import dataclass
from datetime import date
from enum import Enum
from typing import Dict, List

from gelidum import freeze

from ..decorators import immutable


class SemesterHalf(str, Enum):
    """Gemini typically schedules programs for two semesters per year, namely A and B.
    For other observatories, this logic might have to be substantially changed.

    Members:
        - A = 'A'
        - B = 'B'
    """
    A = 'A'
    B = 'B'

    def start_day(self) -> int: # noqa
        return 1

    def end_day(self) -> int: # noqa
        return 31

    def start_month(self) -> int:
        return _semester_half_months[self][0]

    def end_month(self) -> int:
        return _semester_half_months[self][-1]

    @staticmethod
    def determine_half(month: int) -> 'SemesterHalf':
        """
        Given a month, return the SemesterHalf this falls in.

        Args:
            month: the month of the date

        Returns:
            The SemesterHalf in which this month occurs.
        """
        if month < 1 or month > 12:
            raise ValueError(f'Illegal month: {month}')
        for semester_half, months in _semester_half_months.items():
            if month in months:
                return semester_half
        raise ValueError(f'Month {month} cannot be found in any SemesterHalf.')


# Make a FrozenDict with values FrozenList, so the lists cannot be changed, nor can the dict.
_semester_half_months: Dict[SemesterHalf, List[int]] = freeze({
    SemesterHalf.A: [2, 3, 4, 5, 6, 7],
    SemesterHalf.B: [8, 9, 10, 11, 12, 1]
})


@immutable
@dataclass(frozen=True, order=True)
class Semester:
    """ A semester is a period for which programs may be submitted to Gemini.

    Attributes:
        year (int): A four digit year.
        half (SemesterHalf): Two semesters during each year, indicated by the SemesterHalf.
    """
    year: int
    half: SemesterHalf

    def start_date(self) -> date:
        """
        Returns:
            The start date of the Semester.
        """
        return date(year=self.year, month=self.half.start_month(), day=self.half.start_day())

    def end_date(self) -> date:
        """
        Returns:
            The end date of the Semester.
        """
        # Check if the semester spans two years.
        if _semester_half_months[self.half][-1] < _semester_half_months[self.half][0]:
            end_year = self.year + 1
        else:
            end_year = self.year
        return date(year=end_year, month=self.half.end_month(), day=self.half.end_day())

    @staticmethod
    def find_semester_from_date(lookup_date: date) -> 'Semester':
        """
        Given a date, return the Semester in which it occurs.

        Args:
            A date object representing the date.

        Returns:
            The Semester in which this date occurs.
        """
        # Determine which semester half the date falls in.
        semester_half = SemesterHalf.determine_half(lookup_date.month)

        # If the month in the semester half is less than the first entry, it spans over a year.
        if lookup_date.month < _semester_half_months[semester_half][0]:
            semester_year = lookup_date.year - 1
        else:
            semester_year = lookup_date.year

        return Semester(year=semester_year, half=semester_half)

    def __str__(self):
        return f'{self.year}{self.half.value}'
